"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
import pystache
import numpy as np

class Flatten(Layer):

    def __init__(self, idx:int, size:int, input_shape:int, data_format:str):
       
        super().__init__()
        self.idx = idx
        self.size = size
        self.input_shape = input_shape
        self.data_format = data_format
        self.name = 'Flatten'

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Flatten (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Flatten (size must be int)")
        if any(type(shape) != int for shape in self.input_shape[1:]):
            raise TypeError("Error: input_shape in Flatten (all dim must be int)")
        if type(self.data_format) != str:
            raise TypeError("Error: data format type in Flatten")

        ### Checking value consistency ###
        if self.size != self.input_shape[1]*self.input_shape[2]*self.input_shape[3]:
            raise ValueError("Error: size value in Flatten ("+str(self.size)+"!="+str(self.input_shape[1]*self.input_shape[2]*self.input_shape[3])+")")
        if self.data_format not in ['channels_last','channels_first']:
            raise ValueError("Error: data format value in Flatten ("+self.data_format+")")

    def generate_inference_code_layer(self):

        mustach_hash = {}

        if(self.data_format == 'channels_last'):
            mustach_hash['channels_last'] = True
            mustach_hash['input_channels'] = self.input_shape[1]
            mustach_hash['input_height'] = self.input_shape[2]
            mustach_hash['input_width'] = self.input_shape[3]
            mustach_hash['name'] = self.name
            mustach_hash['idx'] = "{:02d}".format(self.idx)
            mustach_hash['path'] = self.path
            mustach_hash['size'] = self.size

        with open(self.template_path+'layers/template_Flatten.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def forward_path_layer(self, input:np.ndarray):
        if(self.data_format == 'channels_last'):
            return np.transpose(np.reshape(input,self.input_shape[1:]),(1,2,0))
        else:
            return input


"""Conv versioning manager.

*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

from ...code_generator import (
    Conv2D,
    Conv2D_6loops,
    Conv2D_indirect_gemm,
    Conv2D_std_gemm
)

def Conv2D_implementation(
        old_layer: Conv2D,
        version: str
) -> Conv2D:
    """Create the layer associated to the required implementation"""
    implemented = {
        "6loops": Conv2D_6loops_implementation,
        "indirect_gemm": Conv2D_indirect_gemm_implementation,
        "std_gemm": Conv2D_std_gemm_implementation
    }

    if version != "6loops":
        version = version[:-3]
    
    return implemented[version](old_layer)



def Conv2D_6loops_implementation(
        old_layer: Conv2D
) -> Conv2D_6loops:
    """Create a Conv2D_6loops layer using the attributs of old_layer"""
    return Conv2D_6loops(
        idx= old_layer.idx,
        conv_algorithm= old_layer.conv_algorithm,
        size= old_layer.size,
        padding= old_layer.padding,
        strides= old_layer.strides,
        kernel_h= old_layer.kernel_h,
        kernel_w= old_layer.kernel_w,
        dilation_rate= old_layer.dilation_rate,
        nb_filters= old_layer.nb_filters,
        input_shape= old_layer.input_shape,
        output_shape= [1,old_layer.output_channels,old_layer.output_height,old_layer.output_width],
        weights= old_layer.weights,
        biases= old_layer.biases,
        activation_function= old_layer.activation_function
    )

def Conv2D_indirect_gemm_implementation(
        old_layer: Conv2D
) -> Conv2D_indirect_gemm:
    """Create a Conv2D_indirect_gemm layer using the attributs of old_layer"""
    return Conv2D_indirect_gemm(
        idx= old_layer.idx,
        conv_algorithm= old_layer.conv_algorithm,
        size= old_layer.size,
        padding= old_layer.padding,
        strides= old_layer.strides,
        kernel_h= old_layer.kernel_h,
        kernel_w= old_layer.kernel_w,
        dilation_rate= old_layer.dilation_rate,
        nb_filters= old_layer.nb_filters,
        input_shape= old_layer.input_shape,
        output_shape= [1,old_layer.output_channels,old_layer.output_height,old_layer.output_width],
        weights= old_layer.weights,
        biases= old_layer.biases,
        activation_function= old_layer.activation_function
    )

def Conv2D_std_gemm_implementation(
        old_layer: Conv2D
) -> Conv2D_std_gemm:
    """Create a Conv2D_std_gemm layer using the attributs of old_layer"""
    return Conv2D_std_gemm(
        idx= old_layer.idx,
        conv_algorithm= old_layer.conv_algorithm,
        size= old_layer.size,
        padding= old_layer.padding,
        strides= old_layer.strides,
        kernel_h= old_layer.kernel_h,
        kernel_w= old_layer.kernel_w,
        dilation_rate= old_layer.dilation_rate,
        nb_filters= old_layer.nb_filters,
        input_shape= old_layer.input_shape,
        output_shape= [1,old_layer.output_channels,old_layer.output_height,old_layer.output_width],
        weights= old_layer.weights,
        biases= old_layer.biases,
        activation_function= old_layer.activation_function
    )


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Resize import Resize
import numpy as np
import math
import pystache

#The mode Nearest of the Resize layers.
#The value in the new tensor is found by applying an rounding operation
class ResizeNearest(Resize):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mode='nearest'
        self.nearest_mode_mapping = {"round_prefer_floor":self.round_prefer_floor,
                                     "round_prefer_ceil":self.round_prefer_ceil,
                                     "floor":self.floor,
                                     "ceil":self.ceil}
        
        self.nearest_mode_implem_mapping = {"round_prefer_floor":self.round_prefer_floor_implem,
                                            "round_prefer_ceil":self.round_prefer_ceil_implem,
                                            "floor":math.floor,
                                            "ceil":math.ceil}
        
        ####### Checking the instantiation#######

        ### Checking argument type ###
        if type(self.nearest_mode) != str:
            raise TypeError("Error: nearest mode in Resize Nearest (must be string)")

        ### Checking value consistency ###
        if self.nearest_mode not in ['round_prefer_floor','round_prefer_ceil','floor','ceil']:
            raise ValueError("Error: nearest mode value in Resize Nearest ("+self.nearest_mode+")")
    
    #Defining the several method to chose the nearest
    def floor(self, x:str, y:str):
        return x+' = floor('+y+');'
    
    def ceil(self, x:str, y:str):
        return x+' = ceil('+y+');'
    
    def round_prefer_floor(self, x:str, y:str):
        return x+' = floor(ceil(2 * ' + y + ') / 2);'
    
    def round_prefer_floor_implem(self, x:float):
        return math.floor(math.ceil(2*x)/2)
    
    def round_prefer_ceil(self, x:str, y:str):
        return x+' = ceil(floor(2 * ' + y + ') / 2);'
    
    def round_prefer_ceil_implem(self, x:float):
        return math.ceil(math.floor(2*x)/2)
    
    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str(output_str+'[y0 + ' + str(self.input_width) + ' * (x0 + ' + str(self.input_height) + ' * f)]')

        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['coordinate_transformation_mode_x'] = self.coordinate_transformation_mode_mapping[self.coordinate_transformation_mode]('i',2,'x')
        mustach_hash['coordinate_transformation_mode_y'] = self.coordinate_transformation_mode_mapping[self.coordinate_transformation_mode]('j',3,'y')
        mustach_hash['nearest_mode_x'] = self.nearest_mode_mapping[self.nearest_mode]('x0','x')
        mustach_hash['nearest_mode_y'] = self.nearest_mode_mapping[self.nearest_mode]('y0','y')

        if(self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('tensor_temp[j + ' + str(self.output_width) + '*(i + ' + str(self.output_height) + '*f)]',self.idx,'j + ' + str(self.output_width) + '*(i + ' + str(self.output_height) + '*f)')

        with open(self.template_path+'layers/Resize/template_ResizeNearest.c.tpl') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)    
    
    def forward_path_layer(self, input:np.ndarray):
        input = input.reshape(self.input_channels, self.input_height, self.input_width)
        output = np.zeros((self.output_channels,self.output_height,self.output_width))
        for f in range(self.output_channels):
            for i in range(self.output_height):
                for j in range(self.output_width):
                    x = self.coordinate_transformation_mode_implem_mapping[self.coordinate_transformation_mode](i,2)
                    x0 = self.nearest_mode_implem_mapping[self.nearest_mode](x)
                    y = self.coordinate_transformation_mode_implem_mapping[self.coordinate_transformation_mode](j,3)
                    y0 = self.nearest_mode_implem_mapping[self.nearest_mode](y)

                    output[f,i,j] = input[f,x0,y0]
        return output



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from . import layers
from .layers import (
    Add_Bias, Concatenate, Dense, Dot, Flatten, Gather, Gemm, InputLayer, MatMul, 
    Softmax, BatchNormalization, Transpose, GatherElements, Tile,
    Add, Average, Broadcast, Divide, Maximum, Minimum, Multiply, Subtract,
    Conv2D, Conv2D_6loops, Conv2D_gemm, Conv2D_indirect_gemm, Conv2D_std_gemm,
    Pad, Edge_pad, Wrap_pad, Reflect_pad, Constant_Pad,
    Pooling2D, MaxPooling2D, AveragePooling2D,
    Reduce, ReduceSum, ReduceMax, ReduceProd, ReduceMin, ReduceMean,
    Resize, ResizeCubic, ResizeLinear, ResizeNearest
)

from .activation_functions import (
    ActivationFunctions, Linear, Sigmoid, ReLu, TanH, Exponential, Logarithm, Clip, LeakyReLu
)

from .neural_network import CodeGenerator

from .Layer import Layer

__all__ = (
    "CodeGenerator",
    "ActivationFunctions", "Linear", "Sigmoid", "ReLu", "TanH", "Exponential", "Logarithm", "Clip", "LeakyReLu",
    "Layer",
    layers.__all__
)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Pooling2D import Pooling2D

import numpy as np

class MaxPooling2D(Pooling2D):
    def __init__(self, **kwds):
        super().__init__(**kwds)
        
        self.name = 'MaxPooling2D'
        self.pooling_function = np.amax
        self.local_var = 'max'
        self.output_var = self.local_var

    def update_local_vars(self):
        
        s = self.local_var +' = -INFINITY;\n'

        return s

    def specific_function(self, index:str, input_of_layer:str):
        s = 'if ('+input_of_layer+'['+index+'] > '+self.local_var+')\n'
        s += '                                '+self.local_var+' = '+input_of_layer+'['+index+'];\n'
    
        return s



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
from ..activation_functions import ActivationFunctions
import numpy as np
import pystache

#The layer which compute the general matrix multiplication
#input: weight tesnsor W and bias tensor B, input tensor T. The tensor must be of 2D
#data: alpha and beta constante used in the operation, transpo un tuple saying if the tensor T or W must be transposed before the operation
#output: The result of the operation """alpha*T*W + beta*B"""
class Gemm(Layer):
    def __init__(self, idx:int, size:int, alpha:float|int, beta:float|int, transA:bool|int, transB:bool|int, weights:np.ndarray, bias:np.ndarray, input_shape:list, output_shape:list, activation_function:ActivationFunctions):
        super().__init__() 
        self.name = 'Gemm'
        self.idx = idx
        self.size = size
        
        if alpha != 1:
            self.alpha = [alpha]
        else:
            self.alpha = []
        if beta != 1:
            self.beta =  [beta]
        else:
            self.beta = []
            
        self.transpo = (transA,transB)
        self.algo_gemm_mapping = {(0,0):self.write_gemm_nn,
                             (0,1):self.write_gemm_nt,
                             (1,1):self.write_gemm_tt,
                             (1,0):self.write_gemm_tn}
        
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]
        if(input_shape):
            self.input_height = input_shape[2]
            self.input_width = input_shape[3]
        else:
            self.input_height = 1
            self.input_width = 1
        
        self.weights = weights
        self.biases = bias
        self.activation_function = activation_function
        self.nb_weights = self.count_elements_array(self.weights)
        self.nb_biases = self.count_elements_array(self.biases)

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Gemm (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Gemm (size must be int)")
        if  type(self.alpha[0]) != float and type(self.alpha[0]) == int:
            raise TypeError("Error: alpha type in Gemm (alpha must be int or float)")
        if  type(self.beta[0]) != float and type(self.beta[0]) == int:
            raise TypeError("Error: beta type in Gemm (beta must be int or float)")
        if any(type(self.transpo[i]) != int and type(self.transpo[i]) != bool for i in range(2)):
            raise TypeError("Error: transpose type in Gemm (must be boolean or int)")
        if type(self.output_height) != int:
            raise TypeError("Error: output height type in Gemm (must be int)")
        if type(self.output_width) != int:
            raise TypeError("Error: output width type in Gemm (must be int)")
        if type(self.input_height) != int:
            raise TypeError("Error: input height type in Gemm (must be int)")
        if type(self.input_width) != int:
            raise TypeError("Error: input width type in Gemm (must be int)")
        if type(self.weights) != np.ndarray:
            raise TypeError("Error: weights in Gemm (weights must be an numpy array)")
        if type(self.biases) != np.ndarray:
            raise TypeError("Error: biases in Gemm (biases must be an numpy array)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in Gemm (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.size != self.output_height*self.output_width:
            raise ValueError("Error: size value in Gemm ("+str(self.size)+"!="+str(self.output_height*self.output_width)+")")
        shape = self.input_height if self.transpo[0] else self.input_width
        if self.weights.shape[self.transpo[1]] != shape:
            raise ValueError("Error: non consistency between weight shape and input shape in Gemm ("+str(self.weights.shape[self.transpo[1]])+"!="+str(shape)+")")
        shape = self.input_width if self.transpo[0] else self.input_height
        if self.output_height != shape:
            raise ValueError("Error: non consistency between input shape and output shape in Gemm ("+str(self.output_height)+"!="+str(shape)+")")
        if self.output_width != self.weights.shape[1-self.transpo[1]]:
            raise ValueError("Error: non consistency between output shape and weight shape in Gemm ("+str(self.output_width)+"!="+str(self.weights.shape[1-self.transpo[1]])+")")
        if any(self.biases.shape[i] != 1 and self.biases.shape[i] != output_shape[3-i] for i in range(len(self.biases.shape))):
            raise ValueError("Error: biases in Gemm not broadcastable to dim "+str((self.output_height,self.output_width)))

        
    #The various ways to compute the operation: 
    
    #None of the tensor ar transposed
    def write_gemm_nn(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_nn.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def write_gemm_nt(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_nt.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def write_gemm_tn(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_tn.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def write_gemm_tt(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('sum')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output_'+str(self.path),self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_tt.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def forward_path_layer(self, input:np.ndarray):
        input = input.reshape(self.input_height,self.input_width)
        if(self.transpo[0]):
            input = input.transpose()
            
        if(self.transpo[1]):
            self.weights = self.weights.transpose()
        
        return self.activation_function.compute(self.alpha * np.dot(input,self.weights) + self.beta * self.biases)
    
    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['size'] = self.size
        mustach_hash['road'] = self.path

        mustach_hash['patches_size'] = self.output_width*self.output_height
        if self.transpo[0]:
            mustach_hash['gemm_code'] = self.algo_gemm_mapping[self.transpo](self.output_height, self.output_width, self.input_height, 'weights_' + self.name + '_' + str("{:02d}".format(self.idx)), self.previous_layer[0].output_str)
        else:
            mustach_hash['gemm_code'] = self.algo_gemm_mapping[self.transpo](self.output_height, self.output_width, self.input_width, 'weights_' + self.name + '_' + str("{:02d}".format(self.idx)), self.previous_layer[0].output_str)

        with open(self.template_path+'layers/Gemm/template_Gemm.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)


"""Topological sort for Graph.

*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

from acetone_nnet.code_generator.Layer import Layer


def tri_topo(dnn: list):
    """Sort the graph (list of nodes) topologically."""
    paths: dict[int, bool] = {}  # Available paths per id
    # The sorted layer list
    sorted_layers: list[Layer] = []
    # the dict stating which layers need to go in a constant
    dict_cst: dict[int, int] = {}
    for layer in dnn:
        # If the node isn't sorted, we sort it
        if layer.sorted is None:
            parcours_prof_topo(sorted_layers, layer)
    for layer in sorted_layers:
        update_path(layer, paths)
        # FIXME Assess what to_save is supposed to do
        #   to_save(layer, dict_cst)
    # Convert paths
    max_path = max(paths.keys()) + 1
    list_path: list[int] = []
    for path, available in paths.items():
        list_path.append(path)
        list_path.append(1 if available else 0)
    return sorted_layers, list_path, max_path, dict_cst


def parcours_prof_topo(sorted_layers: list[Layer], layer: Layer) -> None:
    """Sort layer in list of sorted layers."""
    # The node is currently being sorted
    layer.sorted = 1
    for nxt_layer in layer.next_layer:
        # The next_layer need to be found and sorted before the current node
        if nxt_layer.sorted is None:
            parcours_prof_topo(sorted_layers, nxt_layer)
    # The node is sorted
    layer.sorted = 0
    # And is added to the list of sorted nodes
    sorted_layers.insert(0, layer)


def allocate_path_to_layer(layer: Layer, paths: dict[int, bool]) -> None:
    """Allocate a new or closed path to layer."""
    # Check if there is a previously closed path available
    for path, available in paths.items():
        if available:
            layer.path = path
            paths[path] = False
            break
    else:
        # If there is no available path, we create a new one
        path = len(paths)
        paths[path] = False
        layer.path = path


def update_path(layer: Layer, paths: dict[int, bool]) -> None:
    """Assign path to layer."""
    if len(layer.previous_layer) == 0:
        # if the layer has no previous one, we create a new path
        allocate_path_to_layer(layer, paths)

    # Every subsequent layer needs to have a path
    given = False
    if len(layer.next_layer) > 0:
        # Allocate the path to the first child, if pathless
        if layer.next_layer[0].path is None:
            layer.next_layer[0].path = layer.path
            given = True
        # Allocate paths to other children
        if len(layer.next_layer) > 1:
            for nxt_layer in layer.next_layer[1:]:
                if len(nxt_layer.previous_layer) == 1 and not given:
                    # Allocate path to child with a single parent, if available
                    nxt_layer.path = layer.path
                    given = True
                elif nxt_layer.path is not None:
                    # if the layer already has a path, we do nothing
                    pass
                else:
                    # in any other case, the next layer receive a new path
                    allocate_path_to_layer(nxt_layer, paths)

    # Close path if not allocated to child
    if not given:
        if layer.path is None:
            msg = "Layer has no path"
            raise AssertionError(msg)
        paths[layer.path] = True


def to_save(layer: Layer, dict_cst: dict[int, int]) -> None:
    """Create the dict {idx_layer:idx_cst} saying if a layer must be stored."""
    for parent in layer.previous_layer:
        if parent in dict_cst:
            # if the previous_layer are in the dict, we add one to the number of next_layer already "taken care of"
            parent.sorted += 1

    if len(layer.next_layer) > 1:
        # if the layer has more than one child, it must be stored.
        if len(dict_cst) == 0:
            dict_cst[layer.idx] = 1  # if the dict is empty, we create the first cst

        else:
            given = False
            # Going through the dict, starting from the end (the opened cst are at the end of the dict)
            for i in range(len(dict_cst) - 1, -1, -1):
                # extracting the layer at the i-th position
                past_layer = list(dict_cst.keys())[i]
                # if the layer is complete, we can re-use the same constant
                if past_layer.sorted == len(past_layer.next_layer):
                    past_layer.sorted = 0
                    dict_cst[layer.idx] = dict_cst[past_layer]
                    given = True
                    break
            if not given:  # if no constant have been attributed, we create a new one
                dict_cst[layer.idx] = list(dict_cst.values())[-1] + 1



"""*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

from itertools import islice

import keras
import numpy as np
from keras import activations
from keras.engine.base_layer import Layer
from keras.engine.functional import Functional
from keras.engine.sequential import Sequential

from ...code_generator import (
    Add,
    Average,
    AveragePooling2D,
    BatchNormalization,
    Concatenate,
    Constant_Pad,
    Conv2D,
    Dense,
    Flatten,
    InputLayer,
    LeakyReLu,
    Linear,
    Maximum,
    MaxPooling2D,
    Minimum,
    Multiply,
    ReLu,
    Sigmoid,
    Softmax,
    Subtract,
    TanH,
)
from ...graph import graph_interpretor


def get_layer_size(keras_layer:Layer):
    size = 1
    if type(keras_layer.output_shape) is list:
        for dim in keras_layer.output_shape[0][1:]:
            size = size*dim
    else:
        for dim in keras_layer.output_shape[1:]:
            size = size*dim
    return size

def get_output_dimensions(output:list, data_format:str):
    if type(output) is list:
        dimensions = output[0]
    else:
        dimensions = output

    if(data_format == "channels_first"):
        return dimensions
    elif(len(dimensions) == 4):
        return (dimensions[0],dimensions[3],dimensions[1],dimensions[2])
    else:
        return dimensions

def get_input_dimensions(input:list, data_format:str):
    if(type(input) == list and len(input) == 1):
        dimensions = input[0]
    else:
        dimensions = input

    if(data_format == "channels_last"):
        if(type(dimensions) == list and len(dimensions[0]) == 4):
            dimensions = [(shape[0],shape[3],shape[1],shape[2]) for shape in dimensions]
        elif(type(dimensions) is not list and len(dimensions) == 4):
            dimensions =  (dimensions[0],dimensions[3],dimensions[1],dimensions[2])

    return np.array(dimensions)

def create_actv_function_obj(kears_activation_obj):
        if kears_activation_obj == activations.sigmoid:
            return Sigmoid()
        elif kears_activation_obj == activations.relu:
            return ReLu()
        elif kears_activation_obj == activations.tanh:
            return TanH()
        elif kears_activation_obj == activations.linear:
            return Linear()
        elif kears_activation_obj == activations.leaky_relu:
            return LeakyReLu(0.2)
        elif kears_activation_obj == activations.softmax:
            return Linear()

def load_keras(file_to_parse: Functional|Sequential|str, conv_algorithm:str, debug:None|str):

    if(type(file_to_parse) == str):
        model = keras.models.load_model(file_to_parse)
    else:
        model = file_to_parse

    input_layer_size = 1
    for i in range(1, len(model.input.shape)): #start in idx 1 cause idx 0 represents batch size, so it's None in inference phase
        input_layer_size = input_layer_size * model.input.shape[i]

    data_format = "channels_first"
    if(hasattr(layer, "data_format") and layer.data_format == "channels_last" for layer in model.layers):
        data_format = "channels_last"

    data_type = model.layers[0].dtype

    if data_type == "float64":
        data_type = "double"
        data_type_py = np.float64

    elif data_type == "float32":
        data_type = "float"
        data_type_py = np.float32

    elif data_type == "int":
        data_type = "long int"
        data_type_py = np.int32



    layers = []

    if model.layers[0].__class__.__name__ == "InputLayer":
        l_temp = InputLayer(idx = 0,
                            size = get_layer_size(model.layers[0]),
                            input_shape = get_input_dimensions(model.layers[0].input_shape, data_format),
                            data_format = data_format)
        start = 1
    else:
        l_temp = InputLayer(idx = 0,
                            size = input_layer_size,
                            input_shape = get_input_dimensions(model.input_shape, data_format),
                            data_format = data_format)
        start = 0

    layers.append(l_temp)

    nb_softmax_layers = 0

    for idx, layer_keras in list(islice(enumerate(model.layers), start, None)):
        add_softmax_layer = False
        idx += 1-start
        idx += nb_softmax_layers

        if hasattr(layer_keras, "activation"):
            if layer_keras.activation == keras.activations.softmax:
                add_softmax_layer = True

        if layer_keras.__class__.__name__ == "Dense":
            weights = data_type_py(layer_keras.get_weights()[0])
            if(len(weights.shape) < 3):
                for i in range(3-len(weights.shape)):
                    weights = np.expand_dims(weights, axis=-1)
            weights = np.moveaxis(weights, 2, 0)
            if(len(weights.shape) < 4):
                weights = np.expand_dims(weights, axis=0)
            biases = data_type_py(layer_keras.get_weights()[1])
            current_layer = Dense(idx = idx,
                                  size = get_layer_size(layer_keras),
                                  weights = weights,
                                  biases = biases,
                                  activation_function = create_actv_function_obj(layer_keras.activation))

        elif layer_keras.__class__.__name__ == "Conv2D":
            weights = data_type_py(layer_keras.get_weights()[0])
            if(len(weights.shape) < 3):
                for i in range(3-len(weights.shape)):
                    weights = np.expand_dims(weights, axis=-1)
            weights = np.moveaxis(weights, 2, 0)
            if(len(weights.shape) < 4):
                weights = np.expand_dims(weights, axis=0)
            biases = data_type_py(layer_keras.get_weights()[1])
            current_layer = Conv2D(algorithm = conv_algorithm,
                                    conv_algorithm = conv_algorithm,
                                    idx = idx,
                                    size = get_layer_size(layer_keras),
                                    padding = layer_keras.padding,
                                    strides = layer_keras.strides[0],
                                    kernel_h = layer_keras.kernel_size[0],
                                    kernel_w = layer_keras.kernel_size[1],
                                    dilation_rate = layer_keras.dilation_rate[0],
                                    nb_filters = layer_keras.filters,
                                    input_shape = get_input_dimensions(layer_keras.input_shape, data_format),
                                    output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                    weights = weights,
                                    biases = biases,
                                    activation_function = create_actv_function_obj(layer_keras.activation))

        elif layer_keras.__class__.__name__ == "AveragePooling2D":
            current_layer = AveragePooling2D(idx = idx,
                                             size = get_layer_size(layer_keras),
                                             padding = layer_keras.padding,
                                             strides = layer_keras.strides[0],
                                             pool_size = layer_keras.pool_size[0],
                                             input_shape = get_input_dimensions(layer_keras.input_shape, data_format),
                                             output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                             activation_function = Linear())

        elif layer_keras.__class__.__name__ == "MaxPooling2D":
            current_layer = MaxPooling2D(idx = idx,
                                         size = get_layer_size(layer_keras),
                                         padding = layer_keras.padding,
                                         strides = layer_keras.strides[0],
                                         pool_size = layer_keras.pool_size[0],
                                         input_shape = get_input_dimensions(layer_keras.input_shape, data_format),
                                         output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                         activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Flatten":
            current_layer = Flatten(idx = idx,
                                    size = get_layer_size(layer_keras),
                                    input_shape = get_input_dimensions(layer_keras.input_shape, data_format),
                                    data_format= data_format)

        elif layer_keras.__class__.__name__ == "Add":
            current_layer = Add(idx = idx,
                                size = get_layer_size(layer_keras),
                                input_shapes = get_input_dimensions(layer_keras.input_shape, data_format),
                                output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Multiply":
            current_layer = Multiply(idx = idx,
                                     size = get_layer_size(layer_keras),
                                     input_shapes = get_input_dimensions(layer_keras.input_shape, data_format),
                                     output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                     activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Subtract":
            current_layer = Subtract(idx = idx,
                                     size = get_layer_size(layer_keras),
                                     input_shapes = get_input_dimensions(layer_keras.input_shape, data_format),
                                     output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                     activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Maximum":
            current_layer = Maximum(idx = idx,
                                    size = get_layer_size(layer_keras),
                                    input_shapes = get_input_dimensions(layer_keras.input_shape, data_format),
                                    output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                    activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Minimum":
            current_layer = Minimum(idx = idx,
                                    size = get_layer_size(layer_keras),
                                    input_shapes = get_input_dimensions(layer_keras.input_shape, data_format),
                                    output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                    activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Average":
            current_layer = Average(idx = idx,
                                    size = get_layer_size(layer_keras),
                                    input_shapes = get_input_dimensions(layer_keras.input_shape, data_format),
                                    output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                    activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Concatenate":
            axis = layer_keras.axis
            if data_format == "channels_last":
                if axis == 3:
                    axis =1
                else:
                    axis = axis + 1
            current_layer = Concatenate(idx = idx,
                                        size = get_layer_size(layer_keras),
                                        axis = axis,
                                        input_shapes = get_input_dimensions(layer_keras.input_shape, data_format),
                                        output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                        activation_function = Linear())

        elif layer_keras.__class__.__name__ == "ZeroPadding2D":
            pads = layer_keras.padding
            if type(pads) == int:
                pad_top, pad_bottom = pads, pads
                pad_left, pad_right = pads, pads
            elif type(pads[0]) == int:
                pad_top, pad_bottom = pads[0], pads[0]
                pad_left, pad_right = pads[1], pads[1]
            else:
                pad_top, pad_bottom = pads[0][0], pads[0][1]
                pad_left, pad_right = pads[1][0], pads[1][1]
            pads = [0,0,pad_top,pad_left,0,0,pad_bottom,pad_right]
            current_layer = Constant_Pad(idx = idx,
                                         size = get_layer_size(layer_keras),
                                         pads = pads,
                                         constant_value = 0,
                                         axes = [],
                                         input_shape = get_input_dimensions(layer_keras.input_shape, data_format),
                                         activation_function = Linear())

        elif layer_keras.__class__.__name__ == "BatchNormalization":
            if layers[-1].name == "Conv2D" and not debug:
                scale = data_type_py(layer_keras.get_weights()[0])
                bias = data_type_py(layer_keras.get_weights()[1])
                mean = data_type_py(layer_keras.get_weights()[2])
                var = data_type_py(layer_keras.get_weights()[3])

                weights = layers[-1].weights
                biases = layers[-1].biases

                for z in range(len(weights[0,0,0,:])):
                    alpha = scale[z]/np.sqrt(var[z] + layer_keras.epsilon)
                    B = bias[z] - (mean[z]*alpha)
                    weights[:,:,:,z] = alpha*weights[:,:,:,z]
                    biases[z] = alpha*biases[z] + B

                layers[-1].weights = weights
                layers[-1].biases = biases

                continue
            else:
                current_layer = BatchNormalization(idx = idx,
                                                   size = get_layer_size(layer_keras),
                                                   input_shape = get_input_dimensions(layer_keras.input_shape, data_format),
                                                   epsilon = layer_keras.epsilon,
                                                   scale = data_type_py(layer_keras.get_weights()[0]),
                                                   biases = data_type_py(layer_keras.get_weights()[1]),
                                                   mean = data_type_py(layer_keras.get_weights()[2]),
                                                   var = data_type_py(layer_keras.get_weights()[3]),
                                                   activation_function = Linear())

        elif layer_keras.__class__.__name__ == "Reshape" or layer_keras.__class__.__name__ == "Dropout":
            continue

        else:
            raise TypeError("Error: layer "+layer_keras.__class__.__name__+" not supported\n")

        if type(model) == keras.Sequential:
            if idx - 1 >= 0:
                current_layer.previous_layer.append(layers[idx-1])
                layers[idx-1].next_layer.append(current_layer)
        elif(type(layer_keras.input) == list):
            for input in layer_keras.input:
                for k in range(len(model.layers)):
                    prev_layer = model.layers[k]
                    if prev_layer.name == input._keras_history.layer.name:
                        current_layer.previous_layer.append(layers[k])
                        layers[k].next_layer.append(current_layer)
                        break
        else:
            for k in range(len(model.layers)):
                prev_layer = model.layers[k]
                if prev_layer.name == layer_keras.input._keras_history.layer.name:
                    current_layer.previous_layer.append(layers[k])
                    layers[k].next_layer.append(current_layer)
                    break


        l_temp = current_layer
        layers.append(l_temp)

        if add_softmax_layer:
            nb_softmax_layers += 1
            current_layer = Softmax(idx = idx+1,
                                    size = l_temp.size,
                                    output_shape = get_output_dimensions(layer_keras.output_shape, data_format),
                                    axis = None)
            l_temp.next_layer.append(current_layer)
            current_layer.previous_layer.append(l_temp)
            l_temp = current_layer
            layers.append(l_temp)

    layers, listRoad, maxRoad, dict_cst = graph_interpretor.tri_topo(layers)
    layers = list(map(lambda x:x.find_output_str(dict_cst), layers))
    print("Finished model initialization.")
    return layers, data_type, data_type_py, data_format, maxRoad, dict_cst



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
import numpy as np
import pystache

class Softmax(Layer):

    def __init__(self, idx:int, size:int, output_shape:list, axis:int|None):
        
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'Softmax'
        self.axis = axis

        if self.size in (output_shape):
            self.one_dimension = True
        else:
            self.output_channels = output_shape[1]
            self.output_height = output_shape[2]
            self.output_width = output_shape[3]
            self.one_dimension = False


        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Softmax (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Softmax (size must be int)")
        if  type(self.axis)!= int and self.axis!=None:
            raise TypeError("Error: axis type in Softmax (axis must be int or None)")
        if not self.one_dimension:
            if type(self.output_channels) != int:
                raise TypeError("Error: output channels type in Softmax (must be int)")
            if type(self.output_height) != int:
                raise TypeError("Error: output height type in Softmax (must be int)")
            if type(self.output_width) != int:
                raise TypeError("Error: output width type in Softmax (must be int)")

        ### Checking value consistency ###
        if not self.one_dimension:
            if self.size != self.output_channels*self.output_height*self.output_width:
                raise ValueError("Error: size value in Softmax ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        if axis not in [1,2,3] and axis != None:
            raise ValueError("Error: axis out of bound in Softmax ("+str(axis)+"for tensor in 4 dimension with first dimension unused)")
        

    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['road'] = self.path
        mustach_hash['output_str'] = output_str

        mustach_hash['1D'] = self.one_dimension

        if self.one_dimension:
            mustach_hash['size'] = self.size
        else:
            mustach_hash['output_channels'] = self.output_channels
            mustach_hash['output_height'] = self.output_height
            mustach_hash['output_width'] = self.output_width

            if self.axis == 1:
                mustach_hash['sum_dimension_1'] = self.output_height
                mustach_hash['sum_dimension_2'] = self.output_width
                mustach_hash['reduced_dimension'] = self.output_channels
                mustach_hash['reduced_position_1'] = 'i + '+str(self.output_width)+'*f'
                mustach_hash['reduced_position_2'] = 'i + '+str(self.output_width)+'*(f + '+str(self.output_height)+'*j)'
                mustach_hash['softmax_indice'] = 'j + '+str(self.output_width)+'*i'

            elif self.axis == 2:
                mustach_hash['sum_dimension_1'] = self.output_channels
                mustach_hash['sum_dimension_2'] = self.output_width
                mustach_hash['reduced_dimension'] = self.output_height
                mustach_hash['reduced_position_1'] = 'i + '+str(self.output_width)+'*f'
                mustach_hash['reduced_position_2'] = 'i + '+str(self.output_width)+'*(j + '+str(self.output_height)+'*f)'
                mustach_hash['softmax_indice'] = 'j + '+str(self.output_width)+'*f'

            elif self.axis == 3:
                mustach_hash['sum_dimension_1'] = self.output_channels
                mustach_hash['sum_dimension_2'] = self.output_height
                mustach_hash['reduced_dimension'] = self.output_width
                mustach_hash['reduced_position_1'] = 'i + '+str(self.output_height)+'*f'
                mustach_hash['reduced_position_2'] = 'j + '+str(self.output_width)+'*(i + '+str(self.output_height)+'*f)'
                mustach_hash['softmax_indice'] = 'i + '+str(self.output_height)+'*f'

        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output_'+str(self.path)+'[j]', self.idx, 'j')

        with open(self.template_path+'layers/template_Softmax.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def forward_path_layer(self, input:np.ndarray):
        if not self.one_dimension:
            input = input.reshape(1,self.output_channels,self.output_height,self.output_width)
        else:
            input = input.flatten()
            self.axis = -1

        exp = np.exp(input, dtype=np.float)
        output = exp/np.sum(exp, keepdims=1, axis=self.axis)

        return output


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import keras
from keras.engine.functional import Functional
from keras.engine.sequential import Sequential
import numpy as np

def extract_node_outputs(model:Sequential|Functional):
    return [layer.output for layer in model.layers[1:]]

def extract_targets_indices(model:Functional|Sequential,outputs_name:list[str]):
    targets_indices = []
    for name in outputs_name:
        for i in range(len(model.layers)):
            if model.layers[i].output[0] == name:
                targets_indices.append(i)
    return targets_indices

def debug_keras(target_model:Sequential|Functional|str, dataset:np.ndarray, debug_target:list=[], to_save:bool = False, path:str = ''):
    # Loading the model
    if(type(target_model) == str): 
        model = keras.models.load_model(target_model)
    else:
        model = target_model

    # Tensor output name and inidce for acetone debug
    if debug_target == []:
        inter_layers = extract_node_outputs(model)
        targets_indices = []
    else:
        inter_layers = debug_target
        targets_indices = extract_targets_indices(model, inter_layers)
    
    # Add an output after each name of inter_layers
    functional = [Functional([model.input],[out]) for out in inter_layers]

    # Saving the model
    if to_save:
        for i in range(len(functional)):
            keras.models.save_model(model, path+"model_"+str(i)+".h5")

    # Model inference
    outputs = [func([dataset]) for func in functional]

    for i in range(len(outputs)):
        outputs[i] = outputs[i].ravel().flatten()

    return model, targets_indices, outputs


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Conv2D import Conv2D
import numpy as np
import pystache

class Conv2D_gemm(Conv2D):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.patches_height = self.input_channels * self.kernel_h * self.kernel_w
        self.patches_width = self.output_height * self.output_width
        self.patches_size = self.patches_height * self.patches_width

        self.conv_algorithm = self.conv_algorithm[-7:]
        self.algo_gemm_mapping = { 'gemm_nn' : self.write_gemm_nn,
                                   'gemm_nt' : self.write_gemm_nt,
                                   'gemm_tn' : self.write_gemm_tn,
                                   'gemm_tt' : self.write_gemm_tt}

    def write_gemm_nn(self, m:int, n:int, k:int, A:str, B:str, C:str, direct:bool):

        mustach_hash = {}

        mustach_hash['direct'] = direct
        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['ldA'] = k
        mustach_hash['B'] = B
        mustach_hash['ldB'] = n
        mustach_hash['C'] = C
        mustach_hash['ldC'] = n
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Conv/template_Conv_gemm_nn.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def write_gemm_nt(self, m:int, n:int, k:int, A:str, B:str, C:str, direct:bool):

        mustach_hash = {}

        mustach_hash['direct'] = direct
        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['ldA'] = k
        mustach_hash['B'] = B
        mustach_hash['ldB'] = n
        mustach_hash['C'] = C
        mustach_hash['ldC'] = n
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Conv/template_Conv_gemm_nt.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def write_gemm_tn(self, m:int, n:int, k:int, A:str, B:str, C:str, direct:bool):

        mustach_hash = {}

        mustach_hash['direct'] = direct
        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['ldA'] = k
        mustach_hash['B'] = B
        mustach_hash['ldB'] = n
        mustach_hash['C'] = C
        mustach_hash['ldC'] = n
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Conv/template_Conv_gemm_tn.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def write_gemm_tt(self, m:int, n:int, k:int, A:str, B:str, C:str, direct:bool):

        mustach_hash = {}

        mustach_hash['direct'] = direct
        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['ldA'] = k
        mustach_hash['B'] = B
        mustach_hash['ldB'] = n
        mustach_hash['C'] = C
        mustach_hash['ldC'] = n
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('sum')
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output_'+str(self.path),self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Conv/template_Conv_gemm_tt.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Add import Add
from .Average import Average
from .Broadcast import Broadcast
from .Divide import Divide
from .Maximum import Maximum
from .Minimum import Minimum
from .Multiply import Multiply
from .Subtract import Subtract

__all__ = (
    "Add", "Average", "Broadcast", "Divide", "Maximum", "Minimum", "Multiply", "Subtract"
)


"""*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

import numpy as np
import pystache

from ..Layer import Layer


class InputLayer(Layer):

    def __init__(self, idx: int, size: int, input_shape: list, data_format: str):

        super().__init__()
        self.idx = idx
        self.size = size
        self.input_shape = input_shape
        if len(self.input_shape) == 4:
            self.output_channels = self.input_shape[1]
            self.output_height = self.input_shape[2]
            self.output_width = self.input_shape[3]
        self.data_format = data_format
        self.name = "Input_layer"

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if type(self.idx) != int:
            raise TypeError("Error: idx type in Input Layer (idx must be int)")
        if type(self.size) != int:
            raise TypeError("Error: size type in Input Layer (size must be int)")
        if any(type(shape) != int for shape in self.input_shape[1:]):
            raise TypeError("Error: input_shape in Input Layer (all dim must be int)")
        if type(self.data_format) != str:
            raise TypeError("Error: data format type in Input Layer")

        ### Checking value consistency ###
        prod = 1
        for shape in self.input_shape:
            if shape != None and shape != 0:
                prod *= shape

        if self.size != prod:
            raise ValueError(
                "Error: size value in Input Layer ("
                + str(self.size)
                + "!="
                + str(prod)
                + ")",
            )
        if self.data_format not in ["channels_last", "channels_first"]:
            raise ValueError(
                "Error: data format value in Input Layer (" + self.data_format + ")",
            )

    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash["name"] = self.name
        mustach_hash["idx"] = f"{self.idx:02d}"
        mustach_hash["road"] = self.path

        if self.data_format == "channels_last" and len(self.input_shape) == 4:
            mustach_hash["channels_last"] = True
            mustach_hash["input_channels"] = self.output_channels
            mustach_hash["input_height"] = self.output_height
            mustach_hash["input_width"] = self.output_width
        else:
            mustach_hash["size"] = self.size

        with open(
            self.template_path + "/layers/template_Input_Layer.c.tpl",
        ) as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def forward_path_layer(self, input: np.ndarray):

        return input



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ...Layer import Layer
from ...activation_functions import ActivationFunctions
import pystache
import numpy as np
from abc import abstractmethod

class Reduce(Layer):
    
    def __init__(self, idx:int, size:int, axis:int, keepdims:int, noop_with_empty_axes:int, input_shape:list, activation_function:ActivationFunctions):
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'Reduce'

        self.reduce_func = ''
    
        self.axes = axis
        self.keepdims = keepdims
        self.noop_with_empty_axes = noop_with_empty_axes

        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]

        if 1 in self.axes or (self.axes == () and not self.noop_with_empty_axes):
            self.output_channels = 1
        else:
            self.output_channels = self.input_channels
        
        if 2 in self.axes or (self.axes == () and not self.noop_with_empty_axes):
            self.output_height = 1
        else:
            self.output_height = self.input_height
        
        if 3 in self.axes or (self.axes == () and not self.noop_with_empty_axes):
            self.output_width = 1
        else:
            self.output_width = self.input_width

        self.activation_function = activation_function

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Reduce (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Reduce (size must be int)")
        if  type(self.keepdims)!= int:
            raise TypeError("Error: keepdims type in Reduce (size must be int)")
        if  type(self.noop_with_empty_axes)!= int:
            raise TypeError("Error: noop with empty axes type in Reduce (size must be int)")
        if any('int' not in type(axe).__name__  for axe in self.axes) or type(self.axes) != tuple:
            raise TypeError("Error: axes type in Reduce (must be tuple[int])")
        if type(self.input_channels) != int:
            raise TypeError("Error: input channels type in Reduce (must be int)")
        if type(self.input_height) != int:
            raise TypeError("Error: input height type in Reduce (must be int)")
        if type(self.input_width) != int:
            raise TypeError("Error: input width type in Reduce (must be int)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in Reduce (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.size != self.output_channels*self.output_height*self.output_width:
            raise ValueError("Error: size value in Reduce ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        for axe in self.axes:
            if axe < 0 or axe >= 4 :
                raise ValueError("Error: axis out of bound in Reduce ("+str(axe)+" for tensor in 4 dimension with first dimension unused)")
    
    @abstractmethod
    def forward_path_layer(self,input:np.ndarray):
        pass

    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['output_str'] = output_str
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        which_dim_to_reduc = ''
        if 1 in self.axes:
            which_dim_to_reduc += 'c'
        if 2 in self.axes:
            which_dim_to_reduc += 'h'
        if 3 in self.axes:
            which_dim_to_reduc += 'w'
        
        if self.reduce_func == 'Max':
            mustach_hash['Max'] = True
        elif self.reduce_func == 'Min':
            mustach_hash['Min'] = True
        else:
            mustach_hash['Other'] = True
            if self.reduce_func == 'Mean':
                mustach_hash['Mean'] = True

        if self.reduce_func == 'Sum' or self.reduce_func == 'Mean':
            mustach_hash['starting_value'] = 0
            mustach_hash['func'] = '+'
        elif self.reduce_func == 'Prod':
            mustach_hash['starting_value'] = 1
            mustach_hash['func'] = '*'

        if len(which_dim_to_reduc) == 0:
            if self.noop_with_empty_axes:
                mustach_hash['none'] = True
                if self.activation_function.name != 'linear':
                    mustach_hash['Activation'] = True
                    mustach_hash['activation_function'] = self.activation_function.write_activation_str('output_'+str(self.path)+'[k]')
            else:
                mustach_hash['all'] = True
                mustach_hash['activation_function'] = self.activation_function.write_activation_str('reduced')
                mustach_hash['size'] = self.input_channels*self.input_height*self.input_width
                if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                    mustach_hash['starting_value'] = output_str + "[0]"

        elif len(which_dim_to_reduc) == 1:
            mustach_hash['one'] = True
            mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[k]')

            if which_dim_to_reduc == 'c':
                mustach_hash['output_dimension_1'] = self.output_height
                mustach_hash['output_dimension_2'] = self.output_width
                mustach_hash['reduced_dimension'] = self.input_channels
                mustach_hash['position_1'] = 'i + '+str(self.output_width)+'*f'
                mustach_hash['position_2'] = 'i + '+str(self.input_width)+'*(f + '+str(self.input_height)+'*j)'
                if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                    mustach_hash['starting_value'] = output_str + "[i + "+str(self.input_width)+"*f]"

            elif which_dim_to_reduc == 'h':
                mustach_hash['output_dimension_1'] = self.output_channels
                mustach_hash['output_dimension_2'] = self.output_width
                mustach_hash['reduced_dimension'] = self.input_height
                mustach_hash['position_1'] = 'i + '+str(self.output_width)+'*f'
                mustach_hash['position_2'] = 'i + '+str(self.input_width)+'*(j + '+str(self.input_height)+'*f)'
                if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                    mustach_hash['starting_value'] = output_str + "[i + "+str(self.input_width*self.input_height)+"*f]"

            elif which_dim_to_reduc == 'w':
                mustach_hash['output_dimension_1'] = self.output_channels
                mustach_hash['output_dimension_2'] = self.output_height
                mustach_hash['reduced_dimension'] = self.input_width
                mustach_hash['position_1'] = 'i + '+str(self.output_height)+'*f'
                mustach_hash['position_2'] = 'j + '+str(self.input_width)+'*(i + '+str(self.input_height)+'*f)'
                if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                    mustach_hash['starting_value'] = output_str + "["+str(self.input_width)+"*(i + "+str(self.input_height)+"*f)]"
            
            

        elif len(which_dim_to_reduc) == 2:
            mustach_hash['two'] = True
            mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[k]')

            if which_dim_to_reduc == 'ch':
                mustach_hash['output_dimension'] = self.output_width
                mustach_hash['reduced_dimension_1'] = self.input_channels
                mustach_hash['reduced_dimension_2'] = self.input_height
                mustach_hash['position'] = 'f + '+str(self.input_width)+'*(j + '+str(self.input_height)+'*i)'
                if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                    mustach_hash['starting_value'] = output_str + "["+str(self.input_width)+"*f]"

            elif which_dim_to_reduc == 'cw':
                mustach_hash['output_dimension'] = self.output_height
                mustach_hash['reduced_dimension_1'] = self.input_channels
                mustach_hash['reduced_dimension_2'] = self.input_width
                mustach_hash['position'] = 'j + '+str(self.input_width)+'*(f + '+str(self.input_height)+'*i)'
                if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                    mustach_hash['starting_value'] = output_str + "["+str(self.input_height)+"*f]"

            elif which_dim_to_reduc == 'hw':
                mustach_hash['output_dimension'] = self.output_channels
                mustach_hash['reduced_dimension_1'] = self.input_height
                mustach_hash['reduced_dimension_2'] = self.input_width
                mustach_hash['position'] = 'j + '+str(self.input_width)+'*(i + '+str(self.input_height)+'*f)'
                if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                    mustach_hash['starting_value'] = output_str + "["+str(self.input_channels)+"*f]"
            
            if self.reduce_func == 'Mean':
                mustach_hash['nb_elements'] = mustach_hash['reduced_dimension_1']*mustach_hash['reduced_dimension_2']

        
        elif len(which_dim_to_reduc) == 3:
            mustach_hash ['all'] = True
            mustach_hash['activation_function'] = self.activation_function.write_activation_str('reduced')
            mustach_hash['size'] = self.input_channels*self.input_height*self.input_width
            if self.reduce_func == 'Max' or self.reduce_func == 'Min':
                mustach_hash['starting_value'] = output_str + "[0]"

        with open(self.template_path+'layers/template_Reduce.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
from ..activation_functions import ActivationFunctions
import numpy as np
import pystache

class Transpose(Layer):

    def __init__(self, idx:int, size:int, input_shape:list, perm:list, activation_function:ActivationFunctions):
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'Transpose'

        self.perm = perm
    
        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]

        self.output_channels = input_shape[self.perm[1]]
        self.output_height = input_shape[self.perm[2]]
        self.output_width = input_shape[self.perm[3]]

        self.activation_function = activation_function

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Transpose (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Transpose (size must be int)")
        if any(type(indice) != int for indice in self.perm):
            raise TypeError("Error: perm type in Transpose (must be str or ints)")
        if type(self.input_channels) != int:
            raise TypeError("Error: input channels type in Transpose (must be int)")
        if type(self.input_height) != int:
            raise TypeError("Error: input height type in Transpose (must be int)")
        if type(self.input_width) != int:
            raise TypeError("Error: input width type in Transpose (must be int)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in Transpose (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.size != self.output_channels*self.output_height*self.output_width:
            raise ValueError("Error: size value in Transpose ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        seen = []
        for indice in self.perm:
            if indice < 0 or indice >= 4:
                raise ValueError("Error: perm out of bound in Transpose ("+str(indice)+"for tensor in 4 dimension with first dimension unused)")
            if indice in seen:
                raise ValueError("Error: non unicity of perm's values in Transpose "+str(self.perm))
            seen.append(indice)
    
    def forward_path_layer(self, input:np.ndarray):
        input = np.reshape(input, (1,self.input_channels, self.input_height, self.input_width))
        return np.transpose(input, self.perm)

    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['size'] = self.size
        mustach_hash['road'] = self.path
        mustach_hash['output_str'] = output_str

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[k]')

        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['input_height'] = self.input_height
        mustach_hash['input_width'] = self.input_width

        indices = ['Batch','f','i','j']
        if self.perm[1:] == [2,3,1]:
            mustach_hash['a'] = indices[self.perm[1]]
            mustach_hash['b'] = indices[self.perm[3]]
            mustach_hash['c'] = indices[self.perm[2]]
        elif self.perm[1:] == [3,1,2]:
            mustach_hash['a'] = indices[self.perm[2]]
            mustach_hash['b'] = indices[self.perm[1]]
            mustach_hash['c'] = indices[self.perm[3]]
        else:
            mustach_hash['a'] = indices[self.perm[3]]
            mustach_hash['b'] = indices[self.perm[2]]
            mustach_hash['c'] = indices[self.perm[1]]

        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output_'+str(self.path)+'[j]', self.idx, 'j')

        with open(self.template_path+'layers/template_Transpose.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import numpy as np

def compare_result(acetone_result:list|np.ndarray, reference_result:list|np.ndarray, targets:list, verbose:bool = False):
    try:
        assert len(acetone_result) == len(reference_result)
    except AssertionError as msg:
        print("Error: both result don't have the same size")

        if verbose:
            print(msg)
        
        return False
    
    correct = True
    count = 0
    first = None
    for i in range(len(acetone_result)):
        print('--------------------------------------------')
        print('Comparing',targets[i])
        try:
            np.testing.assert_allclose(acetone_result[i],reference_result[i],atol=5e-06,rtol=5e-06)
        except AssertionError as msg:
            print("Error: output value of",targets[i],"incorrect")
            correct = False
            count += 1
            if not first:
                first = targets[i]
            if verbose:
                print(msg)
        print('--------------------------------------------')
    
    if verbose:
        print("++++++++++++++++++++++++++++++++++++++++++++")
        print("Total number of error:",count,"/",len(acetone_result))
        print("First error at",first)
        print("++++++++++++++++++++++++++++++++++++++++++++")

    return correct

def extract_outputs_c(path_to_output:str, data_type:str, nb_targets:int):
    list_result = []
    targets = []
    to_transpose = False
    with open(path_to_output,'r') as f:
        for i, line in enumerate(f):
    
            if i%2 == 0:
                line = line[:-1].split(' ')
                targets.append(line[0] + " " + line[1])
                to_transpose = int(line[2])
                if to_transpose:
                    shape = (int(line[3]),int(line[4]),int(line[5]))
            else:
                line = line[:-2].split(' ')
                if data_type == 'int':
                    line = list(map(int,line))
                elif data_type == 'double':
                    line = list(map(float,line)) 
                elif data_type == 'float':
                    line = list(map(np.float32,line))
                line = np.array(line)

                if to_transpose:
                    line = np.reshape(line, shape)
                    line = np.transpose(line, (1,2,0))
                    line = line.flatten()

                list_result.append(line)
    
            if i == 2*nb_targets-1:
                break
    f.close()

    return list_result, targets

def extract_outputs_python(result_python:tuple[list], targets_indice:list):
    outputs = []
    targets = []
    for indice in targets_indice:
        for i in range(len(result_python[1])):
            if indice == int(result_python[1][i][-1]):
                outputs.append(result_python[0][i])
                targets.append(result_python[1][i])

    return outputs, targets

def reorder_outputs_python(outputs_python:list, targets_python:list):
    ordered_outputs = []
    ordered_targets = []

    dictionary = {}
    for i in range(len(targets_python)):
        dictionary[int(targets_python[i].split(" ")[-1])] = (outputs_python[i], targets_python[i])

    for element in sorted(dictionary.items(), key=lambda item:item[0]):
        ordered_outputs.append(element[1][0])
        ordered_targets.append(element[1][1])
    
    return ordered_outputs, ordered_targets


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ...Layer import Layer
from ...activation_functions import ActivationFunctions
import numpy as np
import math
from abc import abstractmethod

class Conv2D(Layer):
    
    def __init__(self, idx:int, conv_algorithm:str, size:int, padding:str|np.ndarray, strides:int, kernel_h:int, kernel_w:int, dilation_rate:int, nb_filters:int, input_shape:list, output_shape:list, weights:np.ndarray, biases:np.ndarray, activation_function:ActivationFunctions):
        super().__init__()
        self.conv_algorithm = conv_algorithm
        self.idx = idx
        self.size = size
        self.name = 'Conv2D'
        self.padding = padding
        self.strides = strides
        self.kernel_h = kernel_h
        self.kernel_w = kernel_w
        self.dilation_rate = dilation_rate
        self.nb_filters = nb_filters
    
        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]

        self.input_shape = [self.input_channels, self.input_height, self.input_width]
        self.output_channels = self.nb_filters

        self.weights = weights
        self.biases = biases
        self.activation_function = activation_function
        self.local_var = 'sum'

        self.nb_weights = self.count_elements_array(self.weights)
        self.nb_biases = self.count_elements_array(self.biases)
        self.pad_right, self.pad_left, self.pad_bottom, self.pad_top = self.compute_padding(self.padding,self.input_height, self.input_width, self.kernel_h, self.kernel_w, self.strides, self.dilation_rate)
    
        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Conv2D (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Conv2D (size must be int)")
        if type(conv_algorithm) != str:
            raise TypeError("Error: conv algorithm type in Conv2D (must be str)")
        if type(self.padding) != str and any(type(pad) != int for pad in self.padding):
            raise TypeError("Error: padding type in Conv2D (must be str or ints)")
        if  type(self.strides)!= int:
            raise TypeError("Error: strides type in Conv2D (must be int)")
        if  type(self.kernel_h)!= int:
            raise TypeError("Error: kernel_h type in Conv2D (must be int)")
        if  type(self.kernel_w)!= int:
            raise TypeError("Error: kernel_w type in Conv2D (must be int)")
        if  type(self.dilation_rate)!= int:
            raise TypeError("Error: dilation_rate type in Conv2D (must be int)")
        if  type(self.nb_filters)!= int:
            raise TypeError("Error: nb_filters type in Conv2D (must be int)")
        if type(self.input_channels) != int:
            raise TypeError("Error: input channels type in Conv2D (must be int)")
        if type(self.input_height) != int:
            raise TypeError("Error: input height type in Conv2D (must be int)")
        if type(self.input_width) != int:
            raise TypeError("Error: input width type in Conv2D (must be int)")
        if type(self.output_channels) != int:
            raise TypeError("Error: output channels type in Conv2D (must be int)")
        if type(self.output_height) != int:
            raise TypeError("Error: output height type in Conv2D (must be int)")
        if type(self.output_width) != int:
            raise TypeError("Error: output width type in Conv2D (must be int)")
        if type(self.weights) != np.ndarray:
            raise TypeError("Error: weights in Conv2D (weights must be an numpy array)")
        if type(self.biases) != np.ndarray:
            raise TypeError("Error: biases in Conv2D (biases must be an numpy array)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in Conv2D (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.size != self.output_channels*self.output_height*self.output_width:
            raise ValueError("Error: size value in Conv2D ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        if self.weights.shape != (self.input_channels, self.kernel_h, self.kernel_w, self.nb_filters):
            raise ValueError("Error: non consistency between weight shape and operation parameters in Conv2D ("+str(self.weights.shape)+"!="+str((self.input_channels, self.kernel_h, self.kernel_w, self.nb_filters))+")")
        if len(self.biases.shape) != 1 or self.biases.shape[0] != self.nb_filters:
            raise ValueError("Error: non consistency between the var shape and the output shape in Conv2D ("+str(self.biases.shape)+"!="+str(self.nb_filters)+")")
        if self.output_height != math.floor((self.input_height + self.pad_bottom + self.pad_top - self.kernel_h - (self.kernel_h - 1)*(self.dilation_rate - 1))/self.strides) + 1:
            raise ValueError("Error: non consistency between the output height and the parameter of the operation in Conv2D ("+str(self.output_height)+"!="+str(math.floor((self.input_height + self.pad_bottom + self.pad_top - self.kernel_h - (self.kernel_h - 1)*(self.dilation_rate - 1))/self.strides) + 1)+")")
        if self.output_width != math.floor((self.input_width + self.pad_left + self.pad_right - self.kernel_w - (self.kernel_w - 1)*(self.dilation_rate - 1))/self.strides) + 1:
            raise ValueError("Error: non consistency between the output width and the parameter of the operation in Conv2D ("+str(self.output_width)+"!="+str(math.floor((self.input_width + self.pad_left + self.pad_right - self.kernel_w - (self.kernel_w - 1)*(self.dilation_rate - 1))/self.strides) + 1)+")")
        if self.conv_algorithm not in ['6loops',
                                       'indirect_gemm_nn','indirect_gemm_tn','indirect_gemm_nt','indirect_gemm_tt',
                                       'std_gemm_nn','std_gemm_tn','std_gemm_nt','std_gemm_']:
            raise ValueError("Error: conv algorithm value in Conv2D ("+self.conv_algorithm+")")

    @abstractmethod
    def generate_inference_code_layer(self):
        pass

    def forward_path_layer(self, input:np.ndarray):
        # Conv for chw
        input = input.reshape(self.input_channels, self.input_height, self.input_width)
        
        output = np.zeros((self.nb_filters, self.output_height, self.output_width))
        print(self.weights.shape)

        if self.pad_right and self.pad_left and self.pad_top and self.pad_bottom:
            input_padded = np.zeros((self.input_channels, self.input_height + self.pad_top + self.pad_bottom, self.input_width + self.pad_left + self.pad_right))
            input_padded[:, self.pad_top:-self.pad_bottom, self.pad_left:-self.pad_right] = input
        else:
            input_padded = input
            
        for f in range(self.nb_filters):
            for i in range(self.output_height):
                for j in range(self.output_width):
                        output[f,i,j]=np.sum(input_padded[:, i*self.strides:i*self.strides+self.kernel_h, j*self.strides:j*self.strides+self.kernel_w] 
                                            * self.weights[:,:,:,f]) + self.biases[f]
        return self.activation_function.compute(output)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import numpy as np
from abc import abstractmethod

class ActivationFunctions():
    def __init__(self):
        self.name = ''
        self.comment = ''

    @abstractmethod
    def compute(self, z:np.ndarray):
        pass

    @abstractmethod
    def write_activation_str(self, local_var:str):
        pass

class Sigmoid(ActivationFunctions):
    
    def __init__(self):
        super().__init__()
        self.name = 'sigmoid'
        self.comment = ' and apply sigmoid function'
        #self.layer_type

    def compute(self, z:np.ndarray):
        return 1/(1+np.exp(-z))

    def write_activation_str(self, local_var:str):

        s = '1 / (1 + exp(-'+ local_var +'))'
        
        return s

class ReLu(ActivationFunctions):
    
    def __init__(self):
        super().__init__()
        self.name = 'relu'
        self.comment = ' and apply rectifier'
    
    def compute(self, z:np.ndarray):
        return np.maximum(0,z)

    def write_activation_str(self, local_var:str):

        s = local_var +' > 0 ? '+ local_var +' : 0' # output = condition ? value_if_true : value_if_false
        
        return s

class LeakyReLu(ActivationFunctions):
    
    def __init__(self, alpha:float):
        super().__init__()
        self.name = 'leakyrelu'
        self.comment = ' and apply rectifier'
        self.alpha = alpha

        ### Checking value consistency ###
        if self.alpha < 0:
            raise ValueError("Error: alpha value in LeakyRelu (alpha < 0)")
    
    def compute(self, z:np.ndarray):
        temp_tensor = z.flatten()
        for i in range(len(temp_tensor)):
            if(temp_tensor[i]<0):
                temp_tensor[i] =  self.alpha*temp_tensor[i]
        return temp_tensor.reshape(z.shape)

    def write_activation_str(self, local_var:str):

        s = local_var +' > 0 ? '+ local_var +' : '+str(self.alpha)+'*'+local_var # output = condition ? value_if_true : value_if_false
        
        return s

class TanH(ActivationFunctions):
    def __init__(self):
        super().__init__()
        self.name = 'hyperb_tan'
        self.comment = ' and apply hyperbolic tangent function'
        
    
    def compute(self, z:np.ndarray):
        return (np.exp(z)-np.exp(-z))/(np.exp(z)+np.exp(-z))

    def write_activation_str(self, local_var:str):

        s = '(exp('+ local_var +')-exp(-'+ local_var +'))/(exp('+ local_var +')+exp(-'+ local_var +'))'
        
        return s

class Linear(ActivationFunctions):
    def __init__(self):
        super().__init__()
        self.name = 'linear'
        self.comment = ''
    
    def compute(self, z:np.ndarray):
        return z

    def write_activation_str(self, local_var:str):

        s = local_var
        
        return s
    
class Exponential(ActivationFunctions):
    def __init__(self):
        super().__init__()
        self.name = 'Exponential'
        self.comment = ' and apply exponential function'
    
    def compute(self, z:np.ndarray):
        return np.exp(z)
    
    def write_activation_str(self, local_var:str):

        s = 'exp('+local_var+')'

        return s
    
class Logarithm(ActivationFunctions):
    def __init__(self):
        super().__init__()
        self.name = 'Logarithm'
        self.comment = ' and apply logarithm function'
    
    def compute(self, z:np.ndarray):
        return np.log(z)
    
    def write_activation_str(self, local_var:str):
        
        s = 'log('+local_var+')'
        
        return s

class Clip(ActivationFunctions):
    def __init__(self, max:float|int, min:float|int):
        super().__init__()
        self.name = 'Clip'
        self.comment = ' and apply rectifier'
        self.max = max
        self.min = min

        ### Checking value consistency ###
        if  self.min > self.max:
            raise ValueError("Error: min and max values in Clip ("+str(self.min)+" > "+str(self.max)+")")
    
    def compute(self, z:np.ndarray):
        return np.clip(z,self.min,self.max)
    
    def write_activation_str(self,local_var):
        s = local_var +' > '+str(self.max)+' ? '+ str(self.max) +' : (' + local_var + ' < ' + str(self.min) + ' ? ' + str(self.min) + ' : ' + local_var + ')'
        return s


"""Generic layer versioning manager.

*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

from ..code_generator import Layer
from version_implementation.conv_implementation import Conv2D_implementation


def versioning(
        layers: list[Layer],
        version : dict[int,str]
) -> list[Layer]:
    """Check layers and change the layer version if needed"""
    implemented = {
        "Conv2D": Conv2D_implementation
    }

    keys = list(version.keys())
    for i in keys:
        layer = layers[i]
        layers[i] = implemented[layer.name](layer,version[i])
    
    return layers




"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Resize import Resize
import numpy as np
import math
import pystache

#The mode Linear of the Resize layers
#The value in the output tensor are found thanks to a (bi)linear interpolation
class ResizeLinear(Resize):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mode = 'linear'
        self.template_dict = {'1D':self.template_path+'layers/Resize/template_ResizeLinear1D.c.tpl',
                              '2D':self.template_path+'layers/Resize/template_ResizeLinear2D.c.tpl'}
        
    def generate_inference_code_layer(self):

        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['output_str'] = output_str
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        if(self.activation_function.name != 'linear'):
            mustach_hash['linear'] = True
            mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[j + ' + str(self.output_width) + '*(i + ' + str(self.output_height) + '*f)]')

        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        
        if ((self.input_height == 1) and (self.input_width > 1)):
            mustach_hash['len_dimension'] = self.input_width-1
            mustach_hash['dimension'] = self.input_width
            mustach_hash['coordinate_transformation_mode'] = self.coordinate_transformation_mode_mapping[self.coordinate_transformation_mode]('j',3,'x')
            dimension = '1D'
            #return self.write_cst_interpolation_1D_width()
        elif((self.input_height > 1) and (self.input_width == 1)):
            mustach_hash['len_dimension'] = self.input_height-1
            mustach_hash['dimension'] = self.input_height
            mustach_hash['coordinate_transformation_mode'] = self.coordinate_transformation_mode_mapping[self.coordinate_transformation_mode]('i',2,'x')
            dimension = '1D'
            #return self.write_cst_interpolation_1D_height()
        elif((self.input_height > 1) and (self.input_width > 1)):
            mustach_hash['len_height'] = self.input_height-1
            mustach_hash['len_width'] = self.input_width-1
            mustach_hash['input_width'] = self.input_width
            mustach_hash['input_height'] = self.input_height
            mustach_hash['coordinate_transformation_mode_x'] = self.coordinate_transformation_mode_mapping[self.coordinate_transformation_mode]('i',2,'x')
            mustach_hash['coordinate_transformation_mode_y'] = self.coordinate_transformation_mode_mapping[self.coordinate_transformation_mode]('j',3,'y')
            dimension = '2D'
            #return self.write_cst_interpolation_2D()

        if(self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('tensor_temp[j + ' + str(self.output_width) + '*(i + ' + str(self.output_height) + '*f)]',self.idx,'j + ' + str(self.output_width) + '*(i + ' + str(self.output_height) + '*f)')

        with open(self.template_dict[dimension],'r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template,mustach_hash)

    def forward_path_layer(self, input:np.ndarray):
        input = input.reshape(self.input_channels, self.input_height, self.input_width)
        output = np.zeros((self.output_channels,self.output_height,self.output_width))
        for f in range(self.output_channels):
            for i in range(self.output_height):
                for j in range(self.output_width):
                    x = self.coordinate_transformation_mode_implem_mapping[self.coordinate_transformation_mode](i,2)
                    x = np.clip(x,0,self.input_height-1)
                    y = self.coordinate_transformation_mode_implem_mapping[self.coordinate_transformation_mode](j,3)
                    y = np.clip(y,0,self.input_width-1)

                    x0,x1 = math.floor(x), math.floor(x) + 1
                    y0,y1 = math.floor(y), math.floor(y) + 1

                    f11 = input[f,x0,y0]
                    if x1 >= self.input_height:
                        f21 = 0
                    else:
                        f21 = input[f,x1,y0]
                    if x1 >= self.input_height or y1 >= self.input_width:
                        f22 = 0
                    else:
                        f22 = input[f,x1,y1]
                    if y1 >= self.input_width:
                        f12 = 0
                    else:
                        f12 = input[f,x0,y1]
                    
                    output[f,i,j] = (f11*(x1 - x)*(y1 - y) + f21*(x - x0)*(y1 - y) + f12*(x1 - x)*(y - y0) + f22*(x - x0)*(y - y0))/((x1 - x0)*(y1 - y0))
        
        return output



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Conv2D_gemm import Conv2D_gemm

import pystache

class Conv2D_std_gemm(Conv2D_gemm):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.algo_patch_building_mapping = { 'gemm_nn' : self.write_im2col,
                                             'gemm_nt' : self.write_im2row,
                                             'gemm_tn' : self.write_im2col,
                                             'gemm_tt' : self.write_im2row}

    def write_im2col(self):
        output_str = self.previous_layer[0].output_str
        if('output' in output_str):
            output_str = 'tensor_temp'
        
        mustach_hash = {}

        mustach_hash['patches_height'] = self.patches_height
        mustach_hash['kernel_w'] = self.kernel_w
        mustach_hash['kernel_h'] = self.kernel_h
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['strides'] = self.strides
        mustach_hash['pad_top'] = self.pad_top
        mustach_hash['pad_left'] = self.pad_left
        mustach_hash['input_height'] = self.input_height
        mustach_hash['input_width'] = self.input_width
        mustach_hash['road'] = self.path
        mustach_hash['patches_width'] = self.patches_width
        mustach_hash['output_str'] = output_str

        with open(self.template_path+'layers/Conv/template_im2col.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def write_im2row(self):

        output_str = self.previous_layer[0].output_str
        if('output' in output_str):
            output_str = 'tensor_temp'
        
        mustach_hash = {}

        mustach_hash['patches_height'] = self.patches_height
        mustach_hash['kernel_w'] = self.kernel_w
        mustach_hash['kernel_h'] = self.kernel_h
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['strides'] = self.strides
        mustach_hash['pad_top'] = self.pad_top
        mustach_hash['pad_left'] = self.pad_left
        mustach_hash['input_height'] = self.input_height
        mustach_hash['input_width'] = self.input_width
        mustach_hash['road'] = self.path
        mustach_hash['patches_width'] = self.patches_width
        mustach_hash['output_str'] = output_str

        with open(self.template_path+'layers/Conv/template_im2row.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['size'] = self.size
        mustach_hash['road'] = self.path

        mustach_hash['patch_building_code'] = self.algo_patch_building_mapping[self.conv_algorithm]()
        mustach_hash['patches_size'] = self.nb_filters*self.patches_width
        mustach_hash['gemm_code'] = self.algo_gemm_mapping[self.conv_algorithm](self.nb_filters, self.patches_width, self.patches_height, 'weights_' + self.name + '_' + str("{:02d}".format(self.idx)), "output_"+str(self.path),'tensor_temp',False)

        if('cst' not in self.previous_layer[0].output_str):
            mustach_hash['cst'] = True
            mustach_hash['input_size'] = self.input_channels*self.input_height*self.input_width
        

        with open(self.template_path+'layers/Conv/template_Conv_std_gemm.c.tpl','r') as template_file:
            tempalte = template_file.read()
        template_file.close()

        return pystache.render(tempalte, mustach_hash)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import onnx
import numpy as np
import onnxruntime as rt

def clean_inputs(model:onnx.ModelProto):
    while len(model.graph.input) != 1:
        model.graph.input.pop()

def extract_node_outputs(model:onnx.ModelProto):
    activation = ["Relu","Tanh","Sigmoid","Clip","Exp","Log","LeakyRelu"]
    ignored = ["Dropout"]
    ouputs_name = []
    for node in model.graph.node:
        if node.op_type in activation:
            ouputs_name[-1] = node.output[0]
        elif node.op_type in ignored:
            continue
        else:
            ouputs_name.append(node.output[0])
    return ouputs_name

def extract_targets_indices(model:onnx.ModelProto,outputs_name:list[str]):
    targets_indices = []
    for name in outputs_name:
        for i in range(len(model.graph.node)):
            if model.graph.node[i].output[0] == name:
                targets_indices.append(i)
    return targets_indices

def run_model_onnx(model:onnx.ModelProto, dataset:np.ndarray):
    sess = rt.InferenceSession(model)
    input_name = sess.get_inputs()[0].name
    onnx_result = sess.run(None,{input_name: dataset})

    onnx_result.append(onnx_result.pop(0))
    
    for i in range(len(onnx_result)):
        onnx_result[i] = onnx_result[i].ravel().flatten()
    
    return onnx_result

def debug_onnx(target_model:onnx.ModelProto|str, dataset:np.ndarray, debug_target:list=[], to_save:bool = False, path:str = '', otpimize_inputs:bool = False):
    # Loading the model
    if type(target_model) == str:
        model = onnx.load(target_model)
    else:
        model = target_model

    # Tensor output name and inidce for acetone debug
    if debug_target == []:
        inter_layers = extract_node_outputs(model)
        targets_indices = []
    else:
        inter_layers = debug_target
        targets_indices = extract_targets_indices(model, inter_layers)

    # Add an output after each name of inter_layers
    value_info_protos = []
    shape_info = onnx.shape_inference.infer_shapes(model)
    for idx, node in enumerate(shape_info.graph.value_info):
        if node.name in inter_layers:
            value_info_protos.append(node)

    model.graph.output.extend(value_info_protos)  #  in inference stage, these tensor will be added to output dict.

    # Optimizing the model by removing the useless inputs
    if otpimize_inputs:
        clean_inputs(model)
    
    onnx.checker.check_model(model)

    # Save the model
    if to_save:
        onnx.save(model,path)

    # Model inference
    if type(model) == str:
        run_model = model
    else:
        run_model = model.SerializeToString()
    outputs = run_model_onnx(run_model, dataset)
    
    return model, targets_indices, outputs


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ...Layer import Layer
from ...activation_functions import ActivationFunctions
import numpy as np
import math
import pystache
from abc import abstractmethod

class Pooling2D(Layer):
    def __init__(self, idx:int, size:int, padding:str|np.ndarray, strides:int, pool_size:int, input_shape:list, output_shape:list, activation_function:ActivationFunctions, **kwargs):
        
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = ''
        self.padding = padding
        self.strides = strides
        self.pool_size = pool_size

        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]

        self.output_channels = self.input_channels

        self.pooling_function = ''
        self.local_var = ''
        self.local_var_2 = ''
        self.output_var = ''

        self.activation_function = activation_function

        self.pad_right, self.pad_left, self.pad_bottom, self.pad_top = self.compute_padding(self.padding,self.input_height, self.input_width, self.pool_size,self.pool_size, self.strides)


    ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Pooling (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Pooling (size must be int)")
        if type(self.padding) != str and any(type(pad) != int for pad in self.padding):
            raise TypeError("Error: padding type in Pooling (must be str or ints)")
        if  type(self.strides)!= int:
            raise TypeError("Error: strides type in Pooling (must be int)")
        if  type(self.pool_size)!= int:
            raise TypeError("Error: pool_size type in Pooling (must be int)")
        if type(self.input_channels) != int:
            raise TypeError("Error: input channels type in Pooling (must be int)")
        if type(self.input_height) != int:
            raise TypeError("Error: input height type in Pooling (must be int)")
        if type(self.input_width) != int:
            raise TypeError("Error: input width type in Pooling (must be int)")
        if type(self.output_channels) != int:
            raise TypeError("Error: output channels type in Pooling (must be int)")
        if type(self.output_height) != int:
            raise TypeError("Error: output height type in Pooling (must be int)")
        if type(self.output_width) != int:
            raise TypeError("Error: output width type in Pooling (must be int)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in Pooling (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.size != self.output_channels*self.output_height*self.output_width:
            raise ValueError("Error: size value in Pooling ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        if self.output_height != math.floor((self.input_height + self.pad_bottom + self.pad_top - self.pool_size)/self.strides) + 1:
            raise ValueError("Error: non consistency between the output height and the parameter of the operation in Pooling ("+str(self.output_height)+"!="+str(math.floor((self.input_height + self.pad_bottom + self.pad_top - self.pool_size)/self.strides) + 1)+")")
        if self.output_width != math.floor((self.input_width + self.pad_left + self.pad_right - self.pool_size)/self.strides) + 1:
            raise ValueError("Error: non consistency between the output width and the parameter of the operation in Pooling ("+str(self.output_width)+"!="+str(math.floor((self.input_width + self.pad_left + self.pad_right - self.pool_size)/self.strides) + 1)+")")
    
    @abstractmethod    
    def specific_function(self, index:str, input_of_layer:str):
        pass

    @abstractmethod    
    def update_local_vars(self):
        pass

    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str(self.output_var)

        mustach_hash['input_channels'] = self.input_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['update_local_vars'] = self.update_local_vars()
        mustach_hash['pool_size'] = self.pool_size
        mustach_hash['strides'] = self.strides
        mustach_hash['pad_left'] = self.pad_left
        mustach_hash['pad_top'] = self.pad_top
        mustach_hash['input_height'] = self.input_height
        mustach_hash['input_width'] = self.input_width
        mustach_hash['specific_function'] = self.specific_function('jj + '+str(self.input_width)+'*(ii + '+str(self.input_height)+'*f)', output_str)

        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str(self.output_var,self.idx,'j + '+str(self.output_width)+'*(i + '+str(self.output_height)+'*f)')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True
        
        with open(self.template_path+'layers/template_Pooling2D.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def forward_path_layer(self, input:np.ndarray):
        input = input.reshape(self.input_channels, self.input_height, self.input_width)
        
        output = np.zeros((self.input_channels, self.output_height, self.output_width))
                
        if self.pad_right and self.pad_left and self.pad_top and self.pad_bottom:
            input_padded = np.zeros((self.input_channels, self.input_height + self.pad_top + self.pad_bottom, self.input_width + self.pad_left + self.pad_right))
            input_padded[:, self.pad_top:-self.pad_bottom, self.pad_left:-self.pad_right] = input
        else:
            input_padded = input

        for c in range(self.input_channels):
            for i in range(self.output_height):
                for j in range(self.output_width): 
                    output[c,i,j]= self.pooling_function((input_padded[c, i*self.strides:i*self.strides+self.pool_size, j*self.strides:j*self.strides+self.pool_size]))
        
        return output



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ...Layer import Layer
from ...activation_functions import ActivationFunctions
import numpy as np
from abc import abstractmethod

#The Pad Layers
#Pad alongside each dimmensions
#attribut: the mode of padding required
#input: a tensor to be padded, the desired pads, the value of teh constant if mode == constant
#output:the resized tensor
######################### cf https://onnx.ai/onnx/operators/onnx__Pad.html for the doc
class Pad(Layer):
    
    def __init__(self, idx:int, size:int, pads:np.ndarray, constant_value:float, axes:np.ndarray|list, input_shape:list, activation_function:ActivationFunctions):
        super().__init__()
        self.idx = idx
        self.size = size
        self.pads = pads
        self.constant_value = constant_value
        self.axes = axes
        self.name = 'Pad'
        self.input_shape = input_shape
        self.output_channels = input_shape[1] + pads[1] + pads[5]
        self.output_height = input_shape[2] + pads[2] + pads[6]
        self.output_width = input_shape[3] + pads[3] + pads[7]
        self.mode = ''
        self.activation_function = activation_function

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in Pad (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in Pad (size must be int)")
        if any(type(pad) != int for pad in self.pads):
            raise TypeError("Error: pads type in Pad (must be int)")
        if type(self.constant_value) != float and type(self.constant_value) != int:
            raise TypeError("Error: constant value type in Pad (must be float or int)")
        if type(self.output_channels) != int:
            raise TypeError("Error: output channels type in Pad (must be int)")
        if type(self.output_height) != int:
            raise TypeError("Error: output height type in Pad (must be int)")
        if type(self.output_width) != int:
            raise TypeError("Error: output width type in Pad (must be int)")
        if any(type(shape) != int for shape in self.input_shape[1:]):
            raise TypeError("Error: input shape type in Pad (must be int)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in Pad (activation function must be a sub-classe of acetone_nnet Activation Function)")


        ### Checking value consistency ###
        if self.size != self.output_channels*self.output_height*self.output_width:
            raise ValueError("Error: size value in Pad ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        for axe in self.axes:
            if axe < 0 or axe >= 4:
                raise ValueError("Error: axe out of bound in Pad ("+str(axe)+"for tensor in 4 dimension with first dimension unused)")
    
    def forward_path_layer(self, input:np.ndarray):
        input = input.reshape(self.input_shape[1], self.input_shape[2], self.input_shape[3])
        nb_dim = len(self.pads)//2
        pad_width = [(self.pads[i],self.pads[i+nb_dim]) for i in range(1,nb_dim)] #Constructing the pads accordingly to the numpy nomenclature
        return self.activation_function.compute(np.pad(input,pad_width=pad_width,mode=self.mode,constant_values=self.constant_value,))
    
    @abstractmethod
    def generate_inference_code_layer(self):
        pass


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
from ..activation_functions import ActivationFunctions
import numpy as np
import pystache

class BatchNormalization(Layer):

    def __init__(self, idx:int, size:int, input_shape:list, epsilon:float|int, scale:np.ndarray, biases:np.ndarray, mean:np.ndarray, var:np.ndarray, activation_function:ActivationFunctions):
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'BatchNormalization'
        self.output_channels = input_shape[1]
        self.output_height = input_shape[2]
        self.output_width = input_shape[3]
        self.epsilon = epsilon
        self.scale = scale
        self.mean = mean
        self.var = var
        self.biases = biases
        self.nb_biases = self.count_elements_array(self.biases)

        self.activation_function = activation_function

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in BatchNormalization (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in BatchNormalization (size must be int)")
        if type(self.output_channels) != int:
            raise TypeError("Error: output channels type in BatchNormalization (must be int)")
        if type(self.output_height) != int:
            raise TypeError("Error: output height type in BatchNormalization (must be int)")
        if type(self.output_width) != int:
            raise TypeError("Error: output width type in BatchNormalization (must be int)")
        if  type(self.epsilon) != float and type(self.epsilon) == int:
            raise TypeError("Error: epsilon type in BatchNormalization (epsilon must be int or float)")
        if type(self.scale) != np.ndarray:
            raise TypeError("Error: scale in BatchNormalization (scale must be an numpy array)")
        if type(self.mean) != np.ndarray:
            raise TypeError("Error: mean in BatchNormalization (mean must be an numpy array)")
        if type(self.var) != np.ndarray:
            raise TypeError("Error: var in BatchNormalization (var must be an numpy array)")
        if type(self.biases) != np.ndarray:
            raise TypeError("Error: biases in BatchNormalization (biases must be an numpy array)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in BatchNormalization (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.size != self.output_channels*self.output_height*self.output_width:
            raise ValueError("Error: size value in BatchNormalization ("+str(self.size)+"!="+str(self.output_channels*self.output_height*self.output_width)+")")
        if len(self.scale.shape) != 1 or self.scale.shape[0] != self.output_channels:
            raise ValueError("Error: non consistency between the scale shape and the output shape in BatchNormalization ("+str(self.scale.shape)+"!="+str(self.output_channels)+")")
        if len(self.scale.mean) != 1 or self.mean.shape[0] != self.output_channels:
            raise ValueError("Error: non consistency between the mean shape and the output shape in BatchNormalization ("+str(self.mean.shape)+"!="+str(self.output_channels)+")")
        if len(self.var.shape) != 1 or self.var.shape[0] != self.output_channels:
            raise ValueError("Error: non consistency between the var shape and the output shape in BatchNormalization ("+str(self.var.shape)+"!="+str(self.output_channels)+")")
        if len(self.biases.shape) != 1 or self.biases.shape[0] != self.output_channels:
            raise ValueError("Error: non consistency between the biases shape and the output shape in BatchNormalization ("+str(self.biases.shape)+"!="+str(self.output_channels)+")")

    def generate_inference_code_layer(self):
        #Variable indicating under which name the input tensor is
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['output_str'] = output_str
        mustach_hash['path'] = self.path
        
        if(self.activation_function.name != 'linear'):
            mustach_hash['activation_function'] = self.activation_function.write_activation_str('output_'+str(self.path)+'[k + '+str(self.output_height*self.output_width)+'*f]')

        mustach_hash['input_channels'] = self.output_channels
        mustach_hash['channel_size'] = self.output_height*self.output_width
        mustach_hash['epsilon'] = self.epsilon

        with open(self.template_path+'layers/template_BatchNormalization.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def forward_path_layer(self, input:np.ndarray):
        input = np.reshape(input, (self.output_channels, self.output_height, self.output_width))
        output = []
        for i in range(self.output_channels):
            output.append((input[i] - self.mean[i])/np.sqrt(self.var[i] + self.epsilon)*self.scale[i] + self.biases[i])
        return np.array(output)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
from ..activation_functions import ActivationFunctions
import numpy as np
import pystache

class MatMul(Layer):

    def __init__(self, idx:int, size:int, input_shapes:list, weights:np.ndarray, side:int, activation_function:ActivationFunctions):
        
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'MatMul'
        self.activation_function = activation_function
        self.local_var = 'dotproduct'
        self.side = side
        self.input_shapes = input_shapes

        if weights is not None:
            self.weights = weights
            self.nb_weights = self.count_elements_array(self.weights)
        
        if self.side == 0:
            self.output_channels = self.input_shapes[1]
            self.output_height = self.input_shapes[-2]
            self.output_width = self.weights.shape[-1]
            self.shared_dimension = self.input_shapes[-1]
        elif self.side == 1:
            self.output_channels = self.input_shapes[1]
            self.output_width = self.input_shapes[-1]
            self.output_height = self.weights.shape[-2]
            self.shared_dimension = self.input_shapes[-2]
        elif self.side == 2:
            self.output_channels = self.input_shapes[0][1]
            self.output_height = self.input_shapes[0][-2]
            self.output_width = self.input_shapes[1][-1]
            self.shared_dimension = self.input_shapes[0][-1]

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if  type(self.idx)!= int:
            raise TypeError("Error: idx type in MatMul (idx must be int)")
        if  type(self.size)!= int:
            raise TypeError("Error: size type in MatMul (size must be int)")
        if type(self.input_shapes[0]) == list:
            for input_shape in self.input_shapes:
                if any(type(shape) != int for shape in input_shape):
                    raise TypeError("Error: input_shape in MatMul (all dim must be int)")
        else:
            if any(type(shape) != int for shape in self.input_shapes):
                raise TypeError("Error: input_shape in MatMul (all dim must be int)")
        if hasattr(self, 'weights'):
            if type(self.weights) != np.ndarray:
                raise TypeError("Error: weights in MatMul (weights must be an numpy array)")
        if type(side) != int:
            raise TypeError("Error: side type in MatMul (side must be a boolean)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError("Error: activation function type in MatMul (activation function must be a sub-classe of acetone_nnet Activation Function)")

        ### Checking value consistency ###
        if self.side == 1:
            if self.weights.shape[-1] != self.input_shapes[-2]:
                raise ValueError("Error: non consistency between weight shape and input shape in MatMul ("+str(self.weights.shape[-1])+"!="+str(self.input_shapes[-2])+")")
            if self.size != self.weights.shape[-2] * self.input_shapes[-1]*self.output_channels:
                raise ValueError("Error: size value in MatMul ("+str(self.size)+" !="+str(self.weights.shape[-2] * self.input_shapes[-1])+")")
        elif self.side == 0:
            if self.weights.shape[-2] != self.input_shapes[-1]:
                raise ValueError("Error: non consistency between weight shape and input shape in MatMul ("+str(self.weights.shape[-2])+"!="+str(self.input_shapes[-1])+")")
            if self.size != self.weights.shape[-1] * self.input_shapes[-2]*self.output_channels:
                raise ValueError("Error: size value in MatMul ("+str(self.size)+" !="+str(self.weights.shape[-1] * self.input_shapes[-2])+")")
        elif self.side == 2:
            if self.input_shapes[1][-2] != self.input_shapes[0][-1]:
                raise ValueError("Error: non consistency between weight shape and input shape in MatMul ("+str(self.input_shapes[1][-2])+"!="+str(self.input_shapes[0][-1])+")")
            if self.size != self.input_shapes[1][-1] * self.input_shapes[0][-2]*self.output_channels:
                raise ValueError("Error: size value in MatMul ("+str(self.size)+" !="+str(self.input_shapes[1][-1] * self.input_shapes[0][-2])+")")
        else: 
            raise ValueError("Error: side value in Matmul (0 is Input*Weight, 1 is Weight*Input, 2 is Input_1*Input_2, "+str(self.side)+" is not implemented)")

    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        if(self.activation_function.name != 'linear'):
            mustach_hash['non_linear'] = True
            mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[j + '+str(self.output_width)+'*(i + '+str(self.output_height)+'*f)]')

        mustach_hash['shared_dimension'] = self.shared_dimension
        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width

        if self.side == 0 :
            mustach_hash['output_str_left'] = self.previous_layer[0].output_str
            mustach_hash['output_str_right'] = 'weights_'+self.name+'_'+"{:02d}".format(self.idx)
        elif self.side == 1:
            mustach_hash['output_str_right'] = self.previous_layer[0].output_str
            mustach_hash['output_str_left'] = 'weights_'+self.name+'_'+"{:02d}".format(self.idx)
        elif self.side == 2:
            mustach_hash['output_str_left'] = self.previous_layer[0].output_str
            mustach_hash['output_str_right'] = self.previous_layer[1].output_str


        if(self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str(self.local_var,self.idx,'i')

        
        
        with open(self.template_path+'layers/template_MatMul.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
        
    def forward_path_layer(self, input:np.ndarray|list[np.ndarray]):
        if self.side == 1:
            input_1 = input.reshape(self.input_shapes)
            weights = np.moveaxis(self.weights,3,0)
            weights = np.reshape(weights,(1,1,weights.shape[-1],weights.shape[0]))
            return self.activation_function.compute(np.matmul(weights,input_1))
        elif self.side == 0:
            input_1 = input.reshape(self.input_shapes)
            weights = np.moveaxis(self.weights,3,0)
            weights = np.reshape(weights,(1,1,weights.shape[-1],weights.shape[0]))
            return self.activation_function.compute(np.matmul(input_1,weights))
        elif self.side == 2: 
            input_1 = input[0].reshape(self.input_shapes[0])
            input_2 = input[1].reshape(self.input_shapes[1])
            return self.activation_function.compute(np.matmul(input_1,input_2))




"""*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

import numpy as np
import pystache

from ..activation_functions import ActivationFunctions
from ..Layer import Layer


class Dense(Layer):

    def __init__(
        self,
        idx: int,
        size: int,
        weights: np.ndarray,
        biases: np.ndarray,
        activation_function: ActivationFunctions,
    ):

        super().__init__()
        self.idx = idx
        self.size = size
        self.name = "Dense"
        self.weights = weights
        self.biases = biases
        self.activation_function = activation_function
        self.local_var = "dotproduct"

        self.nb_weights = self.count_elements_array(self.weights)
        self.nb_biases = self.count_elements_array(self.biases)

        ####### Checking the instantiation#######

        ### Checking argument type ###
        if type(self.idx) != int:
            raise TypeError("Error: idx type in Dense (idx must be int)")
        if type(self.size) != int:
            raise TypeError("Error: size type in Dense (size must be int)")
        if type(self.weights) != np.ndarray:
            raise TypeError("Error: weights in Dense (weights must be an numpy array)")
        if type(self.biases) != np.ndarray:
            raise TypeError("Error: biases in Dense (biases must be an numpy array)")
        if not isinstance(self.activation_function, ActivationFunctions):
            raise TypeError(
                "Error: activation function type in Dense (activation function must be a sub-classe of acetone_nnet Activation Function)",
            )

        ### Checking value consistency ###
        if self.size != self.weights.shape[-1]:
            raise ValueError(
                "Error: non consistency between weight shape and output shape in Dense ("
                + str(self.size)
                + "!="
                + str(self.weights.shape[-1])
                + ")",
            )
        if self.size != self.biases.shape[0]:
            raise ValueError(
                "Error: non consistency between biases shape and output shape in Dense ("
                + str(self.size)
                + "!="
                + str(self.weights.shape[-1])
                + ")",
            )

    def generate_inference_code_layer(self):
        # Variable indicating under which name the input tensor is
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash["name"] = self.name
        mustach_hash["idx"] = f"{self.idx:02d}"
        mustach_hash["comment"] = self.activation_function.comment
        mustach_hash["output_str"] = output_str
        mustach_hash["road"] = self.path
        mustach_hash["size"] = self.size

        mustach_hash["activation_function"] = (
            self.activation_function.write_activation_str(self.local_var)
        )

        mustach_hash["prev_size"] = self.previous_layer[0].size

        if self.fused_layer:
            mustach_hash["fused_layer"] = self.fused_layer.write_activation_str(
                self.local_var,
                self.idx,
                "i",
            )

            if self.activation_function.name == "linear":
                mustach_hash["linear"] = True

        with open(self.template_path + "/layers/template_Dense.c.tpl") as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def forward_path_layer(self, input: np.ndarray):
        input = input.reshape(self.previous_layer[0].size)
        return self.activation_function.compute(
            np.dot(input, self.weights) + self.biases,
        )



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .AddBias import Add_Bias
from .Concatenate import Concatenate
from .Dense import Dense
from .Dot import Dot
from .Flatten import Flatten
from .Gather import Gather
from .Gemm import Gemm
from .Input import InputLayer
from .MatMul import MatMul
from .Softmax import Softmax
from .BatchNormalization import BatchNormalization
from .Transpose import Transpose
from .GatherElements import GatherElements
from .Tile import Tile

from . import Broadcast_layers
from .Broadcast_layers import Add, Average, Broadcast, Divide, Maximum, Minimum, Multiply, Subtract 

from . import Conv_layers
from .Conv_layers import Conv2D, Conv2D_6loops, Conv2D_gemm, Conv2D_indirect_gemm, Conv2D_std_gemm

from . import Pad_layers
from .Pad_layers import Pad, Constant_Pad, Edge_pad, Reflect_pad, Wrap_pad

from . import Pooling_layers
from .Pooling_layers import Pooling2D, AveragePooling2D, MaxPooling2D

from . import Reduce_layers
from .Reduce_layers import Reduce, ReduceSum, ReduceMax, ReduceMean, ReduceMin, ReduceProd

from . import Resize_layers
from .Resize_layers import Resize, ResizeCubic, ResizeLinear, ResizeNearest


__all__ = (
    "Add_Bias", "Concatenate", "Dense", "Dot", "Flatten", "Gather", "Gemm", "InputLayer", "MatMul", 
    "Softmax", "BatchNormalization", "Transpose", "GatherElements", "Tile",
    Broadcast_layers.__all__,
    Conv_layers.__all__,
    Pad_layers.__all__,
    Pooling_layers.__all__,
    Reduce_layers.__all__,
    Resize_layers.__all__
)


"""*******************************************************************************
* ACETONE: Predictable programming framework for ML applications in safety-critical systems
* Copyright (c) 2022. ONERA
* This file is part of ACETONE
*
* ACETONE is free software ;
* you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation ;
* either version 3 of  the License, or (at your option) any later version.
*
* ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License along with this program ;
* if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
******************************************************************************
"""

from pathlib import Path

import onnx
from keras.engine.functional import Functional
from keras.engine.sequential import Sequential

from .H5_importer.parser_h5 import load_keras
from .JSON_importer.parser_JSON import load_json
from .NNET_importer.parser_NNET import load_nnet
from .ONNX_importer.parser_ONNX import load_onnx


def parser(file_to_parse: str | Path | onnx.ModelProto | Functional | Sequential, conv_algorithm: str,
            debug: None | str = None, *, normalize: bool = False):
    if isinstance(file_to_parse, str | Path):
        # Retrieve extension for file path
        if isinstance(file_to_parse, str):
            extension = file_to_parse
        elif isinstance(file_to_parse, Path):
            extension = file_to_parse.suffix
        else:
            extension = ""

        if "json" in extension[-4:]:
            # FIXME Path-based functions should take a path or string
            return load_json(str(file_to_parse), conv_algorithm)

        if "onnx" in extension[-4:]:
            # FIXME Path-based functions should take a path or string
            return load_onnx(str(file_to_parse), conv_algorithm, debug)

        if "h5" in extension[-4:]:
            # FIXME Path-based functions should take a path or string
            return load_keras(str(file_to_parse), conv_algorithm, debug)

        if "nnet" in extension[-4:]:
            # FIXME Path-based functions should take a path or string
            return load_nnet(str(file_to_parse), normalize)

        print(f"\nError: model description . {extension[-4:]} not supported")
        raise TypeError("Error: model description ." + extension[-4:] + " not supported\nOnly description supported are: .nnet, .h5, .json, .onnx\n")

    if type(file_to_parse) == onnx.ModelProto:
        return load_onnx(file_to_parse, conv_algorithm, debug)

    if type(file_to_parse) == Functional or type(file_to_parse) == Sequential:
        return load_keras(file_to_parse, conv_algorithm, debug)

    print("\nError: model description .", type(file_to_parse), "not supported")
    raise TypeError("Error: model description .", type(file_to_parse),
                    "not supported\nOnly description supported are: .nnet, .h5, .json, .onnx\n")
